from __future__ import annotations

from collections.abc import Iterable
from functools import cached_property
from types import MappingProxyType
from typing import Any

import numpy as np
import scipy.sparse
import shapely
from matplotlib.axes import Axes
from matplotlib.collections import LineCollection, PathCollection
from matplotlib.quiver import Quiver
from matplotlib.text import Text
from numpy import (
    arange,
    asanyarray,
    concatenate,
    full,
    isscalar,
    ndarray,
    newaxis,
    ones,
    searchsorted,
    zeros,
)
from numpy.linalg import norm
from typing_extensions import Self

from curvey._typing import EdgesLike, PointsLike
from curvey.plot import _get_ax, segments, text
from curvey.triangulation import Triangulation
from curvey.util import _rescale


class Edges:
    """A 'edge soup' of directed line segments defined by their vertex coordinates and connectivity

    Parameters
    ----------
    points
        `(n, 2)` array of vertex coordinates.

    edges
        `(n, 2)` integer array of vertex indices. Can also be `None` for a pure point set.

    point_data
        Point data in key => value format. Values are `(n_points,)` or `(n_points, ndims)`
        arrays.

    edge_data
        Edge data in key => value format. Values are `(n_edges,)` or `(n_points, ndims)`
        arrays.
    """

    def __init__(
        self,
        points: PointsLike,
        edges: EdgesLike | None,
        point_data: dict[str, ndarray] | None = None,
        edge_data: dict[str, ndarray] | None = None,
    ):
        self.points: ndarray = asanyarray(points)
        """`(n, 2)` array of vertex coordinates."""

        self.edges: ndarray = asanyarray(edges) if edges is not None else zeros((0, 2), dtype="int")
        """`(n, 2)` integer array of vertex indices."""

        self._point_data: dict[str, ndarray] = {} if point_data is None else point_data
        self._edge_data: dict[str, ndarray] = {} if edge_data is None else edge_data

    def __repr__(self) -> str:
        items: dict[str, str] = {
            "n_points": str(self.n_points),
            "n_edges": str(self.n_edges),
        }
        if pd := self._point_data:
            items["point_data"] = f"{{{', '.join(pd.keys())}}}"

        if ed := self._edge_data:
            items["edge_data"] = f"{{{', '.join(ed.keys())}}}"

        item_list = ", ".join(f"{k}={v}" for k, v in items.items())
        return f"{self.__class__.__name__}({item_list})"

    @property
    def point_data(self) -> MappingProxyType[str, ndarray]:
        """A read-only view of the point data"""
        return MappingProxyType(self._point_data)

    @property
    def edge_data(self) -> MappingProxyType[str, ndarray]:
        """A read-only view of the edge data"""
        return MappingProxyType(self._edge_data)

    def with_(
        self,
        points: ndarray | None = None,
        edges: ndarray | None = None,
        point_data: dict[str, ndarray] | None = None,
        edge_data: dict[str, ndarray] | None = None,
    ) -> Self:
        """Copy of self replacing some subset of properties"""
        return self.__class__(
            points=self.points if points is None else points,
            edges=self.edges if edges is None else edges,
            point_data=self._point_data if point_data is None else point_data,
            edge_data=self._edge_data if edge_data is None else edge_data,
        )

    def _data_with(
        self,
        n_name: str,
        data: dict[str, ndarray],
        kwargs: dict[str, Any],
    ) -> dict[str, ndarray]:
        """Used for copying e.g. point_data or edge_data with extra data added

        Parameters
        ----------
        n_name
            The expected data size variable, e.g. 'n_points' or 'n_edges'.

        data
             E.g. self.point_data or self.edge_data

        **kwargs
            Data to add in key => value format.

        """
        n = getattr(self, n_name)
        data = data.copy()
        for k, v in kwargs.items():
            if isscalar(v):
                val = full(shape=n, fill_value=v)
            else:
                val = asanyarray(v)
                if val.shape[0] != n:
                    msg = f"Data '{k}' has length {val.shape[0]}, expected {n_name}={n}"
                    raise ValueError(msg)
            data[k] = val
        return data

    def with_point_data(self, **kwargs) -> Self:
        """Attach point data in key=value format

        Values must be `(n_points,)` or `(n_points, n_dims)` arrays, *or* a scalar value, in which
        case the scalar is broadcast to a `(n_points,)` array.
        """
        return self.with_(point_data=self._data_with("n_points", self._point_data, kwargs))

    def with_edge_data(self, **kwargs) -> Self:
        """Attach edge data in key=value format

        Values must be `(n_edges,)` or `(n_edges, n_dims)` arrays, *or* a scalar value, in which
        case the scalar is broadcast to a `(n_edges,)` array.
        """
        return self.with_(edge_data=self._data_with("n_edges", self._edge_data, kwargs))

    def drop_edges(self) -> Self:
        """An `Edges` with only points and point data"""
        return self.with_(edges=zeros((0, 2), dtype="int"), edge_data={})

    def reverse(self) -> Edges:
        """Flip edge direction"""
        return self.with_(edges=self.edges[:, ::-1])

    @property
    def n_points(self) -> int:
        """Number of vertices

        This includes points not referenced by the edges array.
        """
        return len(self.points)

    @cached_property
    def edge_length(self) -> ndarray:
        """A `n_edges` length vector of edge lengths"""
        dedge = self.points[self.edges[:, 1]] - self.points[self.edges[:, 0]]
        return norm(dedge, axis=1)

    @property
    def n_edges(self) -> int:
        """Number of edges"""
        return len(self.edges)

    @classmethod
    def empty(cls) -> Self:
        """An `Edges` with zero points and zero edges"""
        return cls(points=zeros((0, 2)), edges=zeros((0, 2), dtype="int"))

    @classmethod
    def concatenate(cls, *es: Self) -> Self:
        """Concatenate multiple edge sets into one

        Parameters
        ----------
        *es
            Multiple `Edges` to concatenate into a single `Edges`.
        """
        if len(es) == 0:
            return cls.empty()

        if len(es) == 1:
            return es[0]

        idx_offset, points, edges = 0, [], []

        point_keys = set.intersection(*(set(e._point_data.keys()) for e in es))
        point_data: dict[str, list[ndarray]] = {k: [] for k in point_keys}

        edge_keys = set.intersection(*(set(e._edge_data.keys()) for e in es))
        edge_data: dict[str, list[ndarray]] = {k: [] for k in edge_keys}

        for e in es:
            points.append(e.points)
            edges.append(idx_offset + e.edges)

            for k in point_data:
                point_data[k].append(e.point_data[k])

            for k in edge_data:
                edge_data[k].append(e.edge_data[k])

            idx_offset += e.n_points

        return cls(
            points=concatenate(points, axis=0),
            edges=concatenate(edges, axis=0),
            point_data={k: concatenate(v, axis=0) for k, v in point_data.items()},
            edge_data={k: concatenate(v, axis=0) for k, v in edge_data.items()},
        )

    @cached_property
    def shapely(self) -> shapely.MultiLineString:
        """Representation of the edges as a `shapely.MultiLineString"""
        # return shapely.MultiLineString(list(self.points[self.edges]))
        return shapely.multilinestrings(self.points[self.edges])

    @cached_property
    def tree(self) -> shapely.STRtree:
        """A shapely.STRtree of edges for fast distance queries"""
        return shapely.STRtree(self.shapely.geoms)

    def plot_points(
        self,
        color: str | ndarray | Any | None = None,
        size: str | ndarray | float | None = None,
        scale_sz: tuple[float, float] | None = None,
        ax: Axes | None = None,
        **kwargs,
    ) -> PathCollection:
        """Plot a scalar quantity on vertices

        Parameters
        -----------
        color
            If a string, assumed to be a name of a `self.point_data` array. Otherwise, either a
            matplotlib scalar colorlike or length `n` array of scalar vertex
            quantities.

        size
            Name of a `point_data` property, or length `n` scalar vertex quantity to size markers
            by, or a fixed size for all vertices.

        scale_sz
            Min and max sizes to scale the vertex quantity `size` to.

        ax
            Matplotlib axes to plot in. Defaults to the current axes.

        **kwargs
            additional kwargs passed to `matplotlib.pyplot.scatter`

        """
        ax = _get_ax(ax)
        if isinstance(color, str) and (color in self.point_data):
            color = self.point_data[color]

        size = _rescale(size, scale_sz)
        return ax.scatter(self.points[:, 0], self.points[:, 1], s=size, c=color, **kwargs)

    def plot_edges(self, **kwargs) -> LineCollection | Quiver:
        """Plot edges

        See `curvey.plot.segments` for additional kwargs descriptions.
        """
        return segments(
            points=self.points,
            edges=self.edges,
            **kwargs,
        )

    def point_labels(
        self, labels: Iterable[str] | None = None, ax: Axes | None = None, clip=True, **kwargs
    ) -> list[Text]:
        """Draw labels on points"""
        return text(points=self.points, labels=labels, ax=ax, clip=clip, **kwargs)

    def edge_labels(
        self, labels: Iterable[str] | None = None, ax: Axes | None = None, clip=True, **kwargs
    ) -> list[Text]:
        """Draw labels on edge midpoints"""
        midpoints = self.points[self.edges].mean(axis=1)
        return text(points=midpoints, labels=labels, ax=ax, clip=clip, **kwargs)

    def triangulate(
        self,
        max_tri_area: float | None = None,
        min_angle: float | None = None,
        polygon: bool = False,
        holes: ndarray | None = None,
        interior_points: ndarray | None = None,
        extra_params: str | None = None,
    ) -> Triangulation:
        """
        Triangulate the polygon enclosed by the edges with Jonathan Shewchuck's
        `triangle` library.

        The python bindings [triangle](https://rufat.be/triangle/index.html) must be importable.
        They can be installed with `pip install triangle`.

        Note
        ----
        This assumes, but does not enforce, no repeated points. `triangle` will often segfault
        with repeated points.

        Parameters
        ----------
        polygon
            If true, perform constrained polygon triangulation. This is equivalent
            to including `'p'` in `extra_params`.

        max_tri_area
            A global maximum triangle area constraint.

        min_angle
            Minimum angle constraint, in degrees.

        holes
            If this edge set includes edges clockwise bounding an exterior hole, specify a point
            interior to that hole to discard triangles inside that hole.

        interior_points
            Additional vertex constraints in addition to `self.points`

        extra_params
            See the [API documentation](https://rufat.be/triangle/API.html).
            E.g. `extra_params='S10X' specifies a maximum number of 10 Steiner points and suppresses
            exact arithmetic.

        """
        import triangle

        params = ""
        if polygon:
            params += "p"

        if max_tri_area is not None:
            params += f"a{max_tri_area:.17f}"

        if min_angle is not None:
            params += f"q{min_angle:.17f}"

        if extra_params is not None:
            params += extra_params

        if interior_points is None:
            # Pretty sure triangle expects to be able to write into this array?
            points = self.points.copy()
        else:
            points = concatenate([self.points, interior_points])

        constraints = {"vertices": points, "segments": self.edges}
        if holes is not None:
            constraints["holes"] = holes

        d = triangle.triangulate(constraints, params)
        is_boundary = d["vertex_markers"].astype(bool).squeeze()

        out = Triangulation(
            points=d["vertices"],
            faces=d["triangles"],
        )
        out.is_boundary_vertex = is_boundary  # overwrite cached property definition
        return out

    def closest_edge(self, points: PointsLike) -> tuple[ndarray, ndarray]:
        """The edge index and distance to the corresponding closest points to the input

        Parameters
        ----------
        points
            `(n, 2)` array of query points

        Returns
        -------
        edge_idx :
            `(n,)` vector of edge indices

        distance :
            `(n,)` vector of euclidean distances to the closest point on the edge

        """
        points = asanyarray(points)
        if points.ndim == 1:
            points = points[newaxis]

        pts = shapely.MultiPoint(points).geoms
        (_pts_idx, edge_idx), dist = self.tree.query_nearest(
            pts, return_distance=True, all_matches=False
        )

        return edge_idx, dist

    def closest_point(self, points: ndarray):
        """The closest points on the closest edge

        Parameters
        ----------
        points
            `(n, 2)` array of query points

        Returns
        -------
        edge_idx :
            `(n,)` vector of edge indices

        distance :
            `(n,)` vector of euclidean distances to the closest point on the edge

        closest :
            `(n, 2)` vector of the closest point on that edge

        """
        points = asanyarray(points)
        if points.ndim == 1:
            points = points[newaxis]

        pts = shapely.MultiPoint(points).geoms
        (pts_idx, edge_idx), dist = self.tree.query_nearest(
            pts, return_distance=True, all_matches=False
        )

        edges = self.tree.geometries[edge_idx]
        closest = zeros((len(points), 2))
        for i, (e, pt) in enumerate(zip(edges, pts)):
            closest[i, :] = e.interpolate(e.project(pt)).coords[0]

        return edge_idx, dist, closest

    def to_csgraph(self, weighted=True, directed=True) -> scipy.sparse.coo_array:
        """Vertex adjacency array for use with scipy's sparse groph routines

        Parameters
        ----------
        weighted
            If true, edge weights are set to inverse edge lengths. Otherwise, edge weights are
            set to `1`.

        directed
            If true, the adjacency matrix `adj[i, j]` is non-zero only if the
            directed edge `(i, j)` is in `self.edges`. If false, `adj[i, j] == adj[j, i]` is
            non-zero if `(i, j)` or `(j, i)` is in `self.edges`.

        Returns
        -------
        adj :
            A `(n_points, n_points)` adjacency matrix.

        """
        edge_weights = 1 / self.edge_length if weighted else ones(self.n_edges)
        n = self.n_points
        u, v = self.edges.T
        out = scipy.sparse.coo_array((edge_weights, (u, v)), shape=(n, n))
        if directed:
            return out
        return out + out.T

    def drop_degenerate_edges(self) -> Self:
        """Drop edges with zero length"""
        eidx = self.edge_length != 0
        return self.with_(
            edges=self.edges[eidx],
            edge_data={k: v[eidx] for k, v in self._edge_data.items()},
        )

    def drop_unreferenced_verts(self) -> Self:
        """Drop points that aren't referenced by the edge array"""
        orig_verts = arange(self.n_points)
        unq_verts = np.unique(self.edges)
        i = searchsorted(unq_verts, orig_verts)
        is_referenced = orig_verts == unq_verts[i]

        # Keep referenced points
        points = self.points[is_referenced]

        # Remap edges to updated points
        edges = i[self.edges]

        return self.with_(
            points=points,
            edges=edges,
            point_data={k: v[is_referenced] for k, v in self._point_data.items()},
            # Don't need to do anything about edge data
            # edge_data=self._edge_data,
        )



from __future__ import annotations

from collections.abc import Iterator
from functools import cached_property

import networkx as nx
import numpy as np
import shapely
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from numpy import arange, array, asanyarray, concatenate, cross, diff, isin, ndarray, newaxis

import curvey
from curvey._typing import PointsLike, TrisLike
from curvey.plot import _get_ax


class Triangulation:
    """A set of triangles defined by vertex coordinates and connectivity array

    Parameters
    ----------
    points
        `(n_points, 2)` array of vertex coordinates.

    faces
        `(n_faces, 3)` integer array of vertex indices.

    """

    def __init__(self, points: PointsLike, faces: TrisLike):
        self.points: ndarray = asanyarray(points)
        """`(n_points, 2)` array of vertex coordinates."""

        self.faces: ndarray = asanyarray(faces)
        """`(n_faces, 3)` integer array of vertex indices."""

    @property
    def n_points(self) -> int:
        """Number of points"""
        return len(self.points)

    @property
    def n_faces(self) -> int:
        """Number of triangles"""
        return len(self.faces)

    @cached_property
    def edge(self) -> ndarray:
        """A `(n_tris, 2, 2)` array of edge vectors

        `self.edge[i, 0]` and `self.edge[i, 1]` are the unnormalized edge vectors of the first
        two edges in triangle `i`.
        """
        return diff(self.points[self.faces], axis=1)

    def reverse(self) -> Triangulation:
        """Reverse the orientation of each triangle"""
        return Triangulation(points=self.points, faces=self.faces[:, ::-1])

    @cached_property
    def signed_area(self) -> ndarray:
        """A length `n_tris` vector of signed triangle areas"""
        return cross(self.edge[:, 0], self.edge[:, 1]) / 2

    def plot_tris(self, ax: Axes | None = None, **kwargs) -> tuple[Line2D, Line2D]:
        """Plot the triangulation

        Parameters
        ----------
        ax
            Axes to plot in, default current axes.

        **kwargs
            Passed to `matplotlib.pyplot.triplot`
        """
        ax = _get_ax(ax)
        x, y = self.points.T
        return ax.triplot(x, y, self.faces, **kwargs)

    def to_edges(self) -> curvey.edges.Edges:
        """An edge soup

        Returns
        -------
        edges :
            An edge set with points `self.points` and `n_edges=3 * self.n_faces`

        """
        return curvey.edges.Edges(points=self.points, edges=self.edges)

    @cached_property
    def edges(self) -> ndarray:
        """A `(n_faces * 3, 2)` array of edge connectivity"""
        f = self.faces
        return concatenate(
            [
                f[:, [0, 1]],
                f[:, [1, 2]],
                f[:, [2, 0]],
            ],
            axis=0,
        )

    @cached_property
    def is_boundary_vertex(self) -> ndarray:
        """`(n_points,) vector of booleans True if the vertex is on the triangulation boundary"""
        return isin(arange(self.n_points), self.boundary_edges.edges)

    @cached_property
    def shapely(self) -> shapely.MultiPolygon:
        """A `shapely.MultiPolygon` of trangles"""
        return shapely.MultiPolygon([shapely.Polygon(self.points[tri]) for tri in self.faces])

    @cached_property
    def tree(self) -> shapely.STRtree:
        """A `shapely.STRtree` of the triangles"""
        return shapely.STRtree(self.shapely.geoms)

    def signed_distance(self, points: PointsLike) -> ndarray:
        """Signed distance from the boundary

        Signed distance is positive if the point is outside the triangulation, negative inside,
        and zero if the point is on the boundary.
        """
        points = asanyarray(points)
        if points.ndim == 1:
            points = points[newaxis]

        pts = shapely.MultiPoint(points).geoms
        (_pts_idx, _edge_idx), boundary_dist = self.boundary_edges.tree.query_nearest(
            pts, return_distance=True, all_matches=False
        )

        (_pts_idx, _tri_idx), tri_dist = self.tree.query_nearest(
            pts, return_distance=True, all_matches=False
        )
        boundary_dist[tri_dist == 0] *= -1
        return boundary_dist

    @cached_property
    def boundary_edges(self) -> curvey.edges.Edges:
        """Edges on the boundary of the triangulation

        Boundary edges are edges that only belong to one triangle.
        """
        # Border edges only appear once; internal edges twice; once in each orientation
        directed_edges = self.edges
        undirected_edges = np.sort(directed_edges, axis=1)
        _unq_edges, unq_idx, n = np.unique(
            undirected_edges, axis=0, return_index=True, return_counts=True
        )
        directed_border_edges = directed_edges[unq_idx[n == 1]]
        return curvey.edges.Edges(self.points, directed_border_edges)

    def boundary_loops(self, idx_name: str | None = None) -> Iterator[curvey.curve.Curve]:
        """Iterate over directed boundary loops

        Parameters
        ----------
        idx_name
            If supplied, the original point index is stored in the curve metadata parameter
            named `idx`.

        Yields
        ------
        loop :
            Boundary loop as a `curvey.curve.Curve`. Orientation is maintained.

        """
        g = nx.DiGraph()
        g.add_edges_from(self.boundary_edges.edges)
        for loop in nx.simple_cycles(g):
            c = curvey.curve.Curve(self.points[loop])
            if idx_name is not None:
                c = c.with_data(**{idx_name: array(loop)})
            yield c



from __future__ import annotations

import collections
from collections.abc import Iterable
from typing import Any, NamedTuple, cast

from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.collections import LineCollection
from matplotlib.quiver import Quiver
from matplotlib.text import Text
from numpy import asarray, ndarray, newaxis
from numpy.linalg import norm
from numpy.typing import ArrayLike

from curvey.util import _rescale


class _VariableColorSpec(NamedTuple):
    """Interpret a colorspec that could be either a constant color or an array of scalars

    The reason we return both variables instead of an enum is for convenience with e.g.
    matplotlibs quiver, which takes two disjoint arguments, `c` (varied) and `color` (fixed)

    """

    fixed: Any = None
    varied: ArrayLike | None = None

    @property
    def maybe_varied(self) -> tuple[()] | tuple[ArrayLike]:
        # Quiver doesn't allow named parameter for the varied color argument,
        # so we need a little splatting magic
        if self.varied is None:
            return ()

        return (self.varied,)

    @staticmethod
    def parse(
        n_data: int | None,
        supplied: Any,
        default_varied=None,
        default_fixed=None,
    ) -> _VariableColorSpec:
        if supplied is None:
            return _VariableColorSpec(fixed=default_fixed, varied=default_varied)

        if isinstance(supplied, str):  # e.g. color='black'
            return _VariableColorSpec(fixed=supplied, varied=None)

        try:
            supplied_len = len(supplied)
        except TypeError as e:
            msg = "Expected `color` to have a length"
            raise NotImplementedError(msg) from e

        if n_data == supplied_len == 3:
            # This is ambiguous and hopefully rare
            # Could raise a warning or something but whatever
            return _VariableColorSpec(fixed=None, varied=asarray(supplied))

        if n_data == supplied_len:
            return _VariableColorSpec(fixed=None, varied=asarray(supplied))

        msg = f"Expected color to be of length {n_data}, got {supplied_len}"
        raise ValueError(msg)


def _get_ax(ax: Axes | None) -> Axes:
    if ax:
        return ax

    if len(plt.get_fignums()) == 0:
        # making a new axes so don't feel bad about setting some defaults
        _fig, ax = plt.subplots()
        ax = cast(Axes, ax)
        ax.axis("equal")
        return ax

    return plt.gca()


def quiver(
    points: ndarray,
    vectors: ndarray,
    scale: ndarray | None = None,
    color=None,
    scale_length: tuple[float, float] | None = None,
    ax: Axes | None = None,
    **kwargs,
) -> Quiver:
    """Convenience wrapper around `matplotlib.pyplot.quiver`

    Differences:

    - Points and vectors are both (n, 2) arrays instead of four vectors or scalars X, Y, U, V
    - Color specification combines both the fixed scalar kwarg `color` and the generic
      array `C` into a single argument.
    - Angles and units for the vectors are always the same as for the points
    - Axes datalimits are updated to include vector heads.

    Parameters
    ----------
    points
        A `(n, 2)` array of vector heads.

    vectors
        A `(n, 2)` array of vectors.

    scale
        A length `n` vector of length scalars to apply to the vectors.

    color
        Length `n` vector of values to color by, or a
        constant color for all edges.

    scale_length
        Limits to scale vector length to, after applying `scale`.

    ax
        The axes to plot in. Defaults to the current axes.

    **kwargs
        additional kwargs passed to `matplotlib.pyplot.quiver`
    """
    ax = _get_ax(ax)

    points, vectors = asarray(points), asarray(vectors)
    if points.ndim == 1:
        points = points[newaxis]
    if vectors.ndim == 1:
        vectors = vectors[newaxis]

    n = len(points)

    if scale is not None:
        vectors = scale.reshape(-1, 1) * vectors

    if scale_length is not None:
        length = norm(vectors, axis=1, keepdims=True)
        scaled_length = _rescale(length, scale_length)
        vectors = vectors / length * scaled_length

    colorspec = _VariableColorSpec.parse(n, color, default_fixed="black")

    # By default quiver doesn't include vector endpoints in x/y lim calculations
    ax.update_datalim(points + vectors)

    # TODO if width is None, set it to some fraction of the bounding box

    (x, y), (dx, dy) = points.T, vectors.T
    return ax.quiver(
        x,
        y,
        dx,
        dy,
        *colorspec.maybe_varied,
        color=colorspec.fixed,
        angles="xy",
        scale_units="xy",
        scale=1.0,
        **kwargs,
    )


def segments(
    points: ndarray,
    edges: ndarray,
    directed: bool = True,
    color=None,
    width: float | ndarray | None = None,
    scale_width: tuple[float, float] | None = None,
    ax: Axes | None = None,
    **kwargs,
) -> LineCollection | Quiver:
    """Plot line segments supplied as separate points and edge arrays

    Parameters
    ----------
    points
        `(m, 2)` array of vertex coordinates

    edges
        `(n, 2)` integer array of indices into `points` defining the start and end of each segment.

    directed
        If true, plot as a quiver plot, otherwise a line plot.

    color
        The color to plot each edge. Can be a constant color or vector of length `n` specifying
        a color for each edge.

    width
        The thickness of each edge segment, scalar or edge quantity vector.

    scale_width
        Min and max widths to scale the edge quantity to.

    ax
        The matplotlib axes to plot in. Defaults to current axes.

    **kwargs
        Aadditional kwargs passed to `matplotlib.collections.LineCollection`.
    """
    ax = _get_ax(ax)

    if directed:
        if isinstance(width, (collections.abc.Sequence, ndarray)):
            msg = (
                "`width` was supplied as a sequence but "
                "matplotlib.pyplot.Quiver doesn't support scaling individual arrows widths. "
                "Use directed=False to scale edge width."
            )
            raise ValueError(msg)

        return quiver(
            points=points[edges[:, 0]],
            vectors=points[edges[:, 1]] - points[edges[:, 0]],
            color=color,
            ax=ax,
            width=width,
            **kwargs,
        )

    colorspec = _VariableColorSpec.parse(len(edges), supplied=color)
    if colorspec.fixed is not None:
        kwargs["color"] = colorspec.fixed

    width = _rescale(width, scale_width)

    lc = LineCollection(
        segments=points[edges],
        linewidths=width,
        **kwargs,
    )

    if colorspec.varied is not None:
        lc.set_array(colorspec.varied)

    ax.add_collection(lc)

    # Adding a line collection doesn't update limits so do it here
    ax.update_datalim(points)
    ax.autoscale_view()
    return lc


def text(
    points: ndarray,
    labels: Iterable[str] | None = None,
    clip: bool = True,
    ax: Axes | None = None,
    **kwargs,
) -> list[Text]:
    """Draw text at the points

    Parameters
    ----------
    points
        `(n, 2)` array of points.

    labels
        `(n,)` sequence of string labels. If not supplied, defaults to the indices
        `0, 1, ..., n-1`.

    clip
        By default matplotlib doesn't restrict clip labels by axes limits -- it's the default here.

    ax
        Matplotlib axes to draw in. Defaults to current axes.

    **kwargs
        Remaining kwargs passed to `matplotlib.pyplot.text`.

    """
    ax = _get_ax(ax)
    txts: list[Text] = []

    if labels is None:
        labels = (str(i) for i in range(len(points)))

    for (x, y), label in zip(points, labels):
        txt = ax.text(x, y, label, **kwargs)
        txts.append(txt)
        if clip:
            txt.set_clip_on(True)

    return txts



from __future__ import annotations

import pytest
from numpy.testing import assert_array_equal

from curvey.triangulation import Triangulation


@pytest.fixture()
def tri3() -> Triangulation:
    pts = [
        [0, 0],
        [1, 0],
        [0, 1],
    ]
    faces = [[0, 1, 2]]
    return Triangulation(pts, faces)


@pytest.fixture()
def tri4() -> Triangulation:
    pts = [
        [0, 0],
        [1, 0],
        [0.5, 1],
        [0.5, 0.5],
    ]
    faces = [
        [0, 1, 3],
        [1, 2, 3],
        [2, 0, 3],
    ]
    return Triangulation(pts, faces)


def test_edges3(tri3):
    assert_array_equal(
        tri3.edges,
        [
            [0, 1],
            [1, 2],
            [2, 0],
        ],
    )


def test_edges4(tri4):
    assert_array_equal(
        tri4.edges,
        [
            [0, 1],
            [1, 2],
            [2, 0],
            [1, 3],
            [2, 3],
            [0, 3],
            [3, 0],
            [3, 1],
            [3, 2],
        ],
    )


def test_is_boundary_vertex(tri4):
    assert_array_equal(tri4.is_boundary_vertex, [True, True, True, False])


def test_boundary_loops(tri4):
    loops = list(tri4.boundary_loops())
    assert len(loops) == 1
    assert_array_equal(
        loops[0].points,
        [
            [0, 0],
            [1, 0],
            [0.5, 1],
        ],
    )



from __future__ import annotations

import pytest
from numpy.testing import assert_array_equal

from curvey import Curve, Edges


@pytest.fixture()
def tri() -> Edges:
    return Curve([[0, 0], [1, 0], [1, 1]]).to_edges()


@pytest.mark.parametrize(
    ("query", "expected"),
    [
        ([0.1, 0], ([0], [0])),  # Allow single point query
        ([[0.1, 0]], ([0], [0])),
        ([[0.1, 0], [0.2, 0]], ([0, 0], [0, 0])),
        ([[0.5, -1]], ([0], [1])),
        ([[0.5, 0.1]], ([0], [0.1])),
    ],
)
def test_closest_edge(tri, query, expected):
    edge_idx, dist = tri.closest_edge(query)
    expected_edge_idx, expected_dist = expected
    assert_array_equal(edge_idx, expected_edge_idx)
    assert_array_equal(dist, expected_dist)


def test_drop_unreferenced_verts():
    es0 = (
        Edges(
            points=[[0, 0], [0, 1], [1, 1]],
            edges=[[0, 2]],
        )
        .with_point_data(pd=[0, 1, 2])
        .with_edge_data(ed=[999])
    )
    es1 = es0.drop_unreferenced_verts()
    assert_array_equal(es1.points, [[0, 0], [1, 1]])
    assert_array_equal(es1.edges, [[0, 1]])
    assert_array_equal(es1.point_data["pd"], [0, 2])
    assert_array_equal(es1.edge_data["ed"], [999])



from __future__ import annotations

from collections.abc import Sequence
from typing import Union

from numpy import ndarray

PointsLike = Union[
    ndarray,
    Sequence[Sequence[float]],
    Sequence[tuple[float, float]],
]

EdgesLike = Union[
    ndarray,
    Sequence[Sequence[int]],
    Sequence[tuple[int, int]],
]

TrisLike = Union[
    ndarray,
    Sequence[Sequence[int]],
    Sequence[tuple[int, int, int]],
]



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import NamedTuple, cast

import numpy as np
import scipy
from numpy import (
    arccos,
    arcsin,
    argmax,
    array,
    ceil,
    clip,
    concatenate,
    diff,
    linspace,
    ndarray,
    newaxis,
    pi,
    roll,
    sqrt,
    stack,
    where,
    zeros,
)
from numpy.linalg import norm
from sortedcontainers import SortedList

from curvey.edges import Edges
from curvey.util import angle_to_points


@dataclass
class SampledPoints:
    """Points uniformly sampled on the border of a maximal disk on the medial axis"""

    points: ndarray
    """`(n, 2)` array of points"""

    distance: ndarray
    """`(n,)` vector of distance to boundary"""

    direction: ndarray
    """Unit vectors pointing to the closest point on the boundary"""

    def __len__(self) -> int:
        """Number of points"""
        return len(self.points)

    def get_disk(self, i: int) -> Disk:
        """Return the `i`th disk"""
        return Disk(
            pt=self.points[i],
            r=self.distance[i],
            vec=self.direction[i],
        )

    def iter_disks(self) -> Iterator[Disk]:
        """Iterate over each point's inscribed disk"""
        for pt, r, direction in zip(self.points, self.distance, self.direction):
            yield Disk(pt=pt, r=r, vec=direction)

    def biggest_disk(self) -> Disk:
        """The inscribed disk with the largest radius"""
        i = cast(int, argmax(self.distance))
        return self.get_disk(i)

    @staticmethod
    def empty() -> SampledPoints:
        """A `SampledPoints` with no points"""
        return _EMPTY_SAMPLED_PTS


_EMPTY_SAMPLED_PTS = SampledPoints(
    points=zeros((0, 2)),
    distance=zeros(0),
    direction=zeros((0, 2)),
)


class Disk(NamedTuple):
    """A maximally inscribed disk on the medial axis"""

    pt: ndarray
    """Center of the disk"""

    r: float
    """Radius of the disk"""

    vec: ndarray
    """Unit vector pointing towards the boundary"""


class Vertex(NamedTuple):
    """A vertex in the medial axis"""

    disk: Disk
    """Inscribed disk corresponding to the vertex"""

    idx: int
    """Index of this vertex in the MA"""


class VisitedDisks:
    """Keep track of set of all disks in the interior we've already visited while building the MA"""

    def __init__(self):
        # Preallocate storage for disks
        n0 = 100
        self._pts = zeros((n0, 2))
        self._sq_radii = zeros(n0)
        self._i = 0

    def append(self, pt: ndarray, r: float) -> None:
        # Check if we need to increase storage capacity
        n = len(self._pts)
        if self._i == n:
            # Need to allocate more space
            self._pts = concatenate([self._pts, np.zeros_like(self._pts)], axis=0)
            self._sq_radii = concatenate([self._sq_radii, np.zeros_like(self._sq_radii)])

        self._pts[self._i] = pt
        self._sq_radii[self._i] = r**2
        self._i += 1

    def has_visited(self, pts: ndarray) -> ndarray:
        if self._i == 0:
            return zeros(len(pts), dtype=bool)

        dist2 = scipy.spatial.distance.cdist(pts, self._pts[: self._i], "sqeuclidean")
        return (dist2 <= self._sq_radii[newaxis, : self._i]).any(axis=1)

    def discard_visited(self, pts: ndarray) -> ndarray:
        if self._i == 0:
            return pts

        return pts[~self.has_visited(pts)]


class FailedToFindMedialAxis(Exception):
    pass


class ApproxMedialAxisBuilder:
    def __init__(
        self,
        boundary: Edges,
        dist_thresh: float,
        angle_thresh: float,
        abs_err: float,
        pt0: ndarray,
        min_edge_length: float | None = None,
    ):
        self.boundary: Edges = boundary
        self.dist_thresh: float = dist_thresh
        self.angle_thresh: float = angle_thresh
        self.abs_err: float = abs_err
        self.min_edge_length = min_edge_length
        self.verts: list[Vertex] = []
        self.edges: list[tuple[int, int]] = []
        self.visited_disks: VisitedDisks = VisitedDisks()
        self.queue: SortedList[Vertex] = SortedList(key=lambda v: v.disk.r)

        # Initialize a maximal disk on the medial axis
        self.pt0: ndarray = pt0
        disk = self.find_initial_medial_axis_point(pt0)
        v0 = self.add_vertex(disk, prnt=None)
        self.queue.add(v0)

    def find_interior_point(self) -> ndarray:
        tris = self.boundary.triangulate()
        tri = max(tris.shapely.geoms, key=lambda t: t.area)
        return array(tri.centroid.coords[0])

    def find_initial_medial_axis_point(self, pt0: ndarray) -> Disk:
        # pt0 is in the interior
        disk = self.maximal_disk_at(pt0)

        while True:
            sampled = self.sample_disk_boundary(disk)
            if len(sampled) == 0:
                msg = "Failed to find medial axis. The abs_thresh or dist_thresh may be too large."
                raise FailedToFindMedialAxis(msg)

            if ma_pts := self.medial_axis_points(sampled):
                # Found a point on the medial axis
                return ma_pts.biggest_disk()

            # This should converge on the medial axis after a couple of steps
            disk = sampled.biggest_disk()

    def distance_to_boundary(self, points: ndarray) -> SampledPoints:
        _, _, boundary_points = self.boundary.closest_point(points)
        direction = boundary_points - points
        distance = norm(direction, axis=1)
        _distance = distance[:, newaxis]
        # Avoid dividing by zero when the point is exactly on the boundary
        np.divide(direction, _distance, where=_distance > 0, out=direction)
        return SampledPoints(points, distance, direction)

    def maximal_disk_at(self, pt: ndarray) -> Disk:
        return self.distance_to_boundary(pt[newaxis]).get_disk(0)

    def medial_axis_points(self, sampled: SampledPoints) -> SampledPoints:
        """Given a set of points sampled on the boundary of an MA disk, find points on the MA"""

        # Points or their adjacent points that haven't been visited yet
        unvisited = ~self.visited_disks.has_visited(sampled.points)
        unvisited = unvisited | roll(unvisited, -1)

        if not unvisited.any():
            return SampledPoints.empty()

        (i0,) = where(unvisited)
        i1 = (i0 + 1) % len(sampled)

        d0 = sampled.direction[i0]
        d1 = sampled.direction[i1]
        theta = arccos(clip((d0 * d1).sum(axis=1), -1, 1))
        above_angle_thresh = theta > self.angle_thresh
        if not above_angle_thresh.any():
            return SampledPoints.empty()

        d0, d1 = d0[above_angle_thresh], d1[above_angle_thresh]
        p0 = sampled.points[i0[above_angle_thresh]]
        p1 = sampled.points[i1[above_angle_thresh]]

        # Divergence test - heads of the vectors should point in different directions
        dtail2 = ((p1 - p0) ** 2).sum(axis=1)
        dtail = sqrt(dtail2)[:, newaxis]
        head1 = p1 + d1 * dtail
        head0 = p0 + d0 * dtail
        dhead2 = ((head1 - head0) ** 2).sum(axis=1)
        diverges = dhead2 > dtail2

        # noinspection PyUnresolvedReferences
        if not diverges.any():
            return SampledPoints.empty()

        # Return midpoints of adjacent sampled.pts where theta > theta_thresh AND adjacent
        # sampled.direction vectors point in different directions
        ma_pts = (p0[diverges] + p1[diverges]) / 2

        if self.min_edge_length is not None and self.queue:
            # Discard points if they're too close to vertices already in the queue
            queue_points = stack([v.disk.pt for v in self.queue], axis=0)
            dist2 = scipy.spatial.distance.cdist(ma_pts, queue_points, metric="sqeuclidean")
            ma_pts = ma_pts[~(dist2 < self.min_edge_length**2).any(axis=1)]

        if len(ma_pts) == 0:
            return SampledPoints.empty()

        return self.distance_to_boundary(ma_pts)

    def sample_disk_boundary(self, disk: Disk) -> SampledPoints:
        if disk.r < self.abs_err:
            return SampledPoints.empty()

        n = int(ceil(pi / arcsin(self.abs_err / disk.r)))
        n = max(n, 4)
        theta = linspace(0, 2 * pi, n, endpoint=False)
        pts = disk.pt[newaxis] + disk.r * angle_to_points(theta)
        return self.distance_to_boundary(pts)

    def visit(self, pt: ndarray, r: float):
        self.visited_disks.append(pt=pt, r=r)

    def discard_visited(self, pts: ndarray) -> ndarray:
        return self.visited_disks.discard_visited(pts)

    def add_vertex(self, disk: Disk, prnt: int | None) -> Vertex:
        idx = len(self.verts)
        v = Vertex(disk=disk, idx=idx)
        self.verts.append(v)
        if prnt is not None:
            self.edges.append((prnt, idx))
        return v

    def step(self, v0: Vertex):
        # Generate uniformly distributed samples on the disk boundary
        sampled = self.sample_disk_boundary(v0.disk)

        # Neighboring points whose direction vectors to their closest boundary points diverge
        ma_pts = self.medial_axis_points(sampled)

        for disk1 in ma_pts.iter_disks():
            if disk1.r > self.dist_thresh:
                v1 = self.add_vertex(disk1, prnt=v0.idx)
                self.queue.add(v1)
            else:
                self.visit(disk1.pt, disk1.r)

        self.visit(v0.disk.pt, v0.disk.r)

    def run(self, max_step: int | None = None):
        step = 0
        while self.queue:
            step += 1
            vertex = self.queue.pop()
            self.step(vertex)

            if max_step is not None and step == max_step:
                break

    def finalize(self, close_loops=True) -> Edges:
        """Convert the accumulated vertices into a standalone Edges object"""
        if not self.verts:
            return Edges.empty()

        ama = Edges(
            points=stack([v.disk.pt for v in self.verts], axis=0),
            edges=stack(self.edges, axis=0) if self.edges else zeros((0, 2), dtype="int"),
            point_data={"distance": array([v.disk.r for v in self.verts])},
        )

        if close_loops and len(ama.edges):
            ama = self.close_loops(ama)

        return ama

    def close_loops(self, ama: Edges) -> Edges:
        # Look for leaf verts near eachother
        distance = ama.point_data["distance"]
        adj = ama.to_csgraph(weighted=False, directed=False)
        (leaf_verts,) = where(adj.sum(axis=1) == 1)
        if len(leaf_verts) == 0:
            return ama

        leaf_points = ama.points[leaf_verts]
        tree = scipy.spatial.KDTree(leaf_points)
        pairs = tree.query_pairs(distance[leaf_verts].max(), output_type="ndarray")

        # Discard pairs that aren't in eachothers' disks
        dp = diff(leaf_points[pairs], axis=1).squeeze(axis=1)
        dist = norm(dp, axis=1)
        r = distance[pairs].min(axis=1)
        i = dist <= r

        if self.min_edge_length is not None:
            i &= dist >= self.min_edge_length

        leaf_vert_pairs = leaf_verts[pairs[i]]
        edges = concatenate([ama.edges, leaf_vert_pairs], axis=0)
        return ama.with_(edges=edges)



"""A polygon bounded by curves"""

from __future__ import annotations

from collections.abc import Iterable, Iterator
from functools import cached_property, partial
from typing import Protocol, cast

import numpy as np
import shapely
from matplotlib.axes import Axes
from matplotlib.patches import PathPatch
from matplotlib.path import Path
from numpy import argmax, array, ndarray, pi, stack
from typing_extensions import Self

from .approx_medial_axis import ApproxMedialAxisBuilder
from .curve import Curve
from .edges import Edges
from .plot import _get_ax
from .triangulation import Triangulation


class CurveFn(Protocol):
    def __call__(self, _c: Curve, *args, **kwargs) -> Curve: ...


class Polygon:
    """A polygon defined by its boundary curves

    It's assumed that the interior curves have opposite orientation to the exterior,
    but this is not enforced.

    Parameters
    ----------
    exterior
        The exterior boundary `Curve`.

    interiors
        A (possibly) empty sequence of `Curve`s bounding holes in the polygon.

    """

    def __init__(self, exterior: Curve, interiors: Iterable[Curve] | None = None):
        self.exterior: Curve = exterior
        self.interiors: list[Curve] = []
        if interiors is not None:
            self.interiors.extend(interiors)

    def __repr__(self) -> str:
        interiors = ", ".join(repr(c) for c in self.interiors)
        return f"{self.__class__.__name__}(exterior={self.exterior}, interiors=({interiors}))"

    def boundaries(self) -> Iterator[Curve]:
        """Iterate over boundary curves"""
        yield self.exterior
        yield from self.interiors

    @classmethod
    def from_shapely(cls, poly: shapely.Polygon) -> Self:
        """Convert a `shapely.Polygon` to a `curvey.Polygon`"""
        exterior = Curve.from_shapely(poly.exterior)
        interiors = (Curve.from_shapely(c) for c in poly.interiors)
        return cls(exterior=exterior, interiors=interiors)

    @classmethod
    def from_text(cls, text: str, **kwargs) -> list[Self]:
        """Construct a set of polygons from matplotlib's text rendering engine

        ```python
        from curvey import Polygon

        polys = Polygon.from_text("curvey", family="arial", size="18")
        for p in polys:
            p.plot()
        ```

        Parameters
        ----------
        text
            The string to render

        **kwargs
            Remaining kwargs passed to `matplotlib.font_manager.FontProperties`
        """
        from matplotlib.font_manager import FontProperties
        from matplotlib.path import Path
        from matplotlib.textpath import TextToPath

        ttp = TextToPath()
        verts, codes = ttp.get_text_path(FontProperties(**kwargs), text)
        poly_pts: list[ndarray] = cast(list[ndarray], Path(verts, codes).to_polygons())

        # Find interior and exterior boundaries
        # Convert to polygons first so that we can use the 'contains_properly' predicate
        polys = [shapely.Polygon(pts) for pts in poly_pts]
        tree = shapely.STRtree(polys)
        exterior_idxs, interior_idxs = tree.query(polys, "contains_properly")

        # Map interior boundaries to exteriors
        poly_idxs: dict[int, list[int]] = {
            i: [] for i in (set(range(len(polys))) - set(interior_idxs))
        }
        for e_idx, i_idx in zip(exterior_idxs, interior_idxs):
            poly_idxs[e_idx].append(i_idx)

        out: list[Self] = []
        for e_idx, i_idxs in poly_idxs.items():
            exterior = Curve(poly_pts[e_idx]).drop_repeated_points().to_ccw()
            interiors = (Curve(poly_pts[i]).drop_repeated_points().to_cw() for i in i_idxs)
            out.append(cls(exterior, interiors))

        return out

    def to_shapely(self) -> shapely.Polygon:
        """Convert a `curvey.Polygon` to a `shapely.Polygon`"""
        return shapely.Polygon(
            self.exterior.to_shapely("ring"), [c.to_shapely("ring") for c in self.interiors]
        )

    def to_edges(self) -> Edges:
        """An edge soup representation of all the edges in the polygon boundaries"""
        return Edges.concatenate(*(c.to_edges() for c in self.boundaries()))

    def apply(self, fn: CurveFn, *args, **kwargs) -> Self:
        """Apply a curve function to boundary curves

        ```python
        from curvey import Curve, Polygon
        poly = Polygon.from_text("e", family='arial')[0]
        poly = poly.apply(Curve.split_long_edges, thresh=1)
        poly.plot()
        ```

        Parameters
        ----------
        fn
            A function `Curve -> Curve`

        *args
        **kwargs
            Additional arguments passed to the function
        """
        fn = partial(fn, *args, **kwargs)
        exterior = fn(self.exterior)
        interiors = (fn(c) for c in self.interiors)
        return self.with_(exterior=exterior, interiors=interiors)

    def with_(self, exterior: Curve, interiors: Iterable[Curve]) -> Self:
        return self.__class__(exterior=exterior, interiors=interiors)

    @cached_property
    def signed_area(self) -> float:
        """Area enclosed by the polygon

        This is simply equal to the sum of the signed areas enclosed by the polygon boundaries.
        Signed area is positive if the polygon is positively (counter-clockwise) oriented.
        """
        return sum(c.signed_area for c in self.boundaries())

    @cached_property
    def area(self) -> float:
        """Absolute area"""
        return abs(self.signed_area)

    def plot(self, **kwargs):
        """Plot polygon boundary

        All kwargs are passed to [Curve.plot][curvey.curve.Curve.plot].
        """
        for c in self.boundaries():
            c.plot(**kwargs)

    def plot_edges(self, **kwargs):
        """Plot polygon boundary edges

        All kwargs are passed to [Curve.plot_edges][curvey.curve.Curve.plot_edges].
        """
        for c in self.boundaries():
            c.plot_edges(**kwargs)

    def to_orientation(self, orientation: int = 1) -> Self:
        """A polygon with the specified orientation

        Parameters
        ----------
        orientation
            Must be 1 or -1. `orientation=1` is a polygon whose exterior boundary is oriented
            counter-clockwise, and whose internal boundaries (if any) have clockwise orientation.
        """
        return self.with_(
            exterior=self.exterior.to_orientation(orientation),
            interiors=(c.to_orientation(-orientation) for c in self.interiors),
        )

    def to_ccw(self) -> Self:
        """A positively-oriented (counter-clockwise) polygon"""
        return self.to_orientation(1)

    def to_cw(self) -> Self:
        """A negatively-oriented (clockwise) polygon"""
        return self.to_orientation(-1)

    def _iter_hole_points(self) -> Iterator[ndarray]:
        """Yield points inside interior holes"""
        for c in self.interiors:
            tris = c.to_ccw().to_edges().triangulate()
            i = argmax(tris.signed_area)
            centroid = tris.points[tris.faces[i]].mean(axis=0)
            yield centroid

    @cached_property
    def hole_points(self) -> ndarray:
        """A `(len(self.interiors), 2)` array of points inside interior holes

        This is probably only useful for triangulation.
        """
        return stack(list(self._iter_hole_points()), axis=0)

    @cached_property
    def boundary(self) -> Edges:
        """Boundary edges as an edge soup"""
        return Edges.concatenate(*(c.to_edges() for c in self.boundaries()))

    def triangulate(
        self,
        max_tri_area: float | None = None,
        min_angle: float | None = None,
        extra_params: str | None = None,
        **kwargs,
    ) -> Triangulation:
        """Triangulate the polygon

        Parameters
        ----------
        max_tri_area
            A global maximum triangle area constraint.

        min_angle
            Minimum angle constraint, in degrees.

        extra_params
            See the [API documentation](https://rufat.be/triangle/API.html).
            E.g. `extra_params='S10X' specifies a maximum number of 10 Steiner points and suppresses
            exact arithmetic.

        **kwargs
            Remaining kwargs passed to `Edges.triangulate`.

        """
        tris = self.boundary.triangulate(
            polygon=True,
            max_tri_area=max_tri_area,
            min_angle=min_angle,
            holes=self.hole_points,
            extra_params=extra_params,
            **kwargs,
        )
        tris.boundary_edges = self.boundary  # overwrite cached property defn
        return tris

    def approximate_medial_axis(
        self,
        dist_thresh: float,
        abs_err: float,
        angle_thresh: float = pi / 3,
        min_edge_length: float | None = None,
        pt0: ndarray | None = None,
        close_loops: bool = True,
        **kwargs,
    ) -> Edges:
        """Construct the approximate medial axis of the polygon

        Implementation of [*Efficient and Robust Computation of an Approximated Medial Axis.*
        Yuandong Yang, Oliver Brock, and Robert N. Moll. 2004.](
        https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=cfc187181ce85d983843c4a184651dbd2a07e7e5)

        The algorithm operates as follows:

        1. Locate an initial point on the medial axis.
        2. Construct a maximally inscribed disk at that point
        3. Uniformly sample points on the boundary of that disk
        4. For each of the sampled points, construct direction vectors pointing at their
        corresponding closest points on the polygon boundary.
        5. Compare the angles between the direction vectors of adjacent points sampled on the
        disk boundary. If the vectors diverge, i.e. the difference in angle exceeds a threshold
        `angle_thresh`, the disk is assumed to intersect the medial axis at midpoint between those
        two adjacent points.
        6. Points found in the previous step are added to the medial axis, and also added to a queue
        to repeated sample maximally inscribed disks as per steps 2-5.

        Parameters
        ----------
        dist_thresh
            Distance from the boundary to stop propagating the medial axis.

        abs_err
            The error allowed in the MA vertex positions. Smaller numbers sample inscribed disks
            more finely.

        angle_thresh
            Angle discreprancy (in radians) to count as a medial axis intersection. Default
            is $pi / 3$.

        min_edge_length
            Prevent adding new vertices if they're within this distance of other vertices

        pt0
            A arbitrary starting point interior to the polygon to begin searching for the medial
            axis. If not supplied, this is chosen automatically by choosing the centroid of the
            largest triangle of the triangulated polygon.

        close_loops
            The standard AMA algorithm produces medial axes in the form of a tree graph. As a final
            post-processing step, look for pairs of leaf vertices within eachother's disks
            and add edges connecting them.

        Returns
        -------
        ama :
            The approximate medial axis as an `curvey.edge.Edges` object. The distance of each
            vertex in the medial axis from the polygon boundary is stored in the `distance`
            point data property.

        """
        if pt0 is None:
            tris = self.triangulate()
            tri = max(tris.shapely.geoms, key=lambda t: t.area)
            pt0 = array(tri.centroid.coords[0])

        b = ApproxMedialAxisBuilder(
            boundary=self.boundary,
            dist_thresh=dist_thresh,
            angle_thresh=angle_thresh,
            abs_err=abs_err,
            min_edge_length=min_edge_length,
            pt0=pt0,
            **kwargs,
        )
        b.run()
        return b.finalize(close_loops=close_loops)

    def to_matplotlib(self) -> Path:
        """Convert a `Polygon` to a `matplotlib.path.Path`"""
        from matplotlib.path import Path

        paths = [c.to_matplotlib() for c in self.boundaries()]
        vertices = np.concatenate([p.vertices for p in paths], axis=0)
        codes = np.concatenate([p.codes for p in paths])
        return Path(vertices, codes)

    def plot_polygon(self, ax: Axes | None = None) -> PathPatch:
        """Plot a filled polygon"""
        ax = _get_ax(ax)
        patch = PathPatch(self.to_matplotlib())
        ax.add_patch(patch)
        ax.update_datalim(self.exterior.points)
        ax.autoscale_view()
        return patch
